#include "astar_graph.h"

void AStarGraph::connect(Vertex& a, Vertex& b)
{
	int length;

	length = sqrt(pow((b.posX - a.posX), 2) + pow((b.posY - a.posY), 2) + pow((b.posZ - a.posZ), 2));
	if (validate_connection(a, b, length))
	{
		a.connect(b, length);
		b.connect(a, length);
	}
}

bool AStarGraph::validate_connection(Vertex& vert_a, Vertex& vert_b, int length)
{
	if (length <= 30)
	{
		float a, b, c, d;
		int plane_count = 0;
		for (auto plane : planes)
		{
			float t = 0;
			float sumT = 0;
			plane.GetPlaneEquation(a, b, c, d);

			//Calculate if the connection between the two points is blocked by a plane
			sumT = (a *(vert_b.posX - vert_a.posX)) + (b* (vert_b.posY - vert_a.posY)) + (c*(vert_b.posZ - vert_a.posZ));
			if (sumT != 0)
			{
				t = ((d - ((a*vert_a.posX) + (b*vert_a.posY) + (c*vert_a.posZ))) / sumT);
				if (t >= 0 && t <= 1)
				{
					int collision_x, collision_y;
					collision_x = vert_a.posX + (t*(vert_b.posX - vert_a.posX));
					collision_y = vert_a.posY + (t*(vert_b.posY - vert_a.posY));
					if (plane.CheckCollisionPoint(collision_x, collision_y))
					{
						return false;
					}
				}
			}
			plane_count++;
		}

		return true;
	}

	return false;
}

void AStarGraph::display()
{
	for (auto v : graph)
	{
		cout << v->index << ": " << v->distance << endl;
	}

	cout << endl;
}

void AStarGraph::init_graph(int num_vertices, int num_obstacles)
{
	srand(time(NULL));

	int segment = 100 / num_vertices;

	//Add the vertices
	for (int i = 0; i < num_vertices; i++)
	{
		Vertex* vertex = nullptr;
		vertex = new Vertex(i, rand() % 100, rand() % 100, rand() % 100);
		graph.push_back(vertex);
	}

	//Add the plane obstalces
	for (int i = 0; i < num_obstacles; i++)
	{
		planes.push_back(Plane(rand() % 100, rand() % 100, rand() % 100));
	}

	//Attempt to connect the vertices appropriately
	for (auto v : graph)
	{
		for (auto vertex : graph)
		{
			if (vertex->index != v->index)
			{
				connect(*v, *vertex);
			}
		}
	}
}

void AStarGraph::clean_up()
{
	for (auto node : graph)
	{
		delete node;
	}
}

void AStarGraph::display_graph()
{
	for (auto node : graph)
	{
		cout << "Node " << node->index << ":" << endl;
		cout << "	Position: " << node->posX << ", " << node->posY << ", " << node->posZ << endl;
		if (node->neighbours.size() > 0)
		{
			cout << "	Connections: " << endl;
			for (auto neighbour : node->neighbours)
			{
				cout << "		Node: " << neighbour->neighbour->index << " Distance: " << neighbour->length << endl;
			}
		}
		cout << endl;
	}
}

void AStarGraph::test()
{
	int start_index, end_index;
	start_index = rand() % graph.size();
	end_index = rand() % graph.size();

	while (end_index == start_index)
	{
		end_index = rand() % graph.size();
	}

	cout << "\nFinding route from Node " << start_index << " to Node " << end_index << endl;
	find_path(start_index, end_index);
}

void AStarGraph::speed_test()
{
	int start_index, end_index;
	start_index = 0;
	end_index = graph.size() - 1;
	while (end_index == start_index)
	{
		end_index = rand() % graph.size();
	}

	find_path_fast(start_index, end_index);
}

void AStarGraph::find_path_fast(int start_index, int end_index)
{
	Vertex *start = nullptr;
	Vertex *end = nullptr;
	Vertex *current_node = nullptr;

	std::list<int> route_list;

	int distance = 0;
	for (auto v : graph)
	{
		if (v->index == start_index)
		{
			start = v;
		}

		if (v->index == end_index)
		{
			end = v;
		}
	}

	if (!start || !end)
	{
		cout << "Something went wrong." << endl;
		return;
	}

	start->distance = 0;

	for (auto v : graph)
	{
		v->CalculateHCost(*end);
	}

	open_set.push(start);
	
	//Save the completed route
	if (compute_route(end))
	{
		current_node = end;

		int route_length = current_node->distance;

		while (current_node->distance != 0)
		{
			route_list.push_front(current_node->index);

			for (auto v : current_node->neighbours)
			{
				if (v->neighbour->distance == (current_node->distance - v->length))
				{
					current_node = v->neighbour;
				}
			}
		}

		route_list.push_front(start->index);
	}
}

void AStarGraph::find_path(int start_index, int end_index)
{
	Vertex *start = nullptr;
	Vertex *end = nullptr;
	Vertex *current_node = nullptr;

	std::list<int> route_list;

	int distance = 0;
	for (auto v : graph)
	{
		if (v->index == start_index)
		{
			start = v;
		}
		
		if (v->index == end_index)
		{
			end = v;
		}
	}

	if (!start || !end)
	{
		cout << "Something went wrong." << endl;
		return;
	}

	start->distance = 0;

	for (auto v : graph)
	{
		v->CalculateHCost(*end);
	}

	open_set.push(start);

	//Save the completed route to be outputted
	if (compute_route(end))
	{
		current_node = end;

		int route_length = current_node->distance;

		while (current_node->distance != 0)
		{
			route_list.push_front(current_node->index);

			for (auto v : current_node->neighbours)
			{
				if (v->neighbour->distance == (current_node->distance - v->length))
				{
					current_node = v->neighbour;
				}
			}
		}

		cout << endl;

		route_list.push_front(start->index);

		for (auto v : route_list)
		{
			cout << "Node " << v << endl;
		}

		cout << "\nThe route length is " << route_length << "\n\n";
	}
	else
	{
		cout << "There is no path to the target." << endl;
	}
}

bool AStarGraph::compute_route(Vertex* end)
{
	if (open_set.empty())
	{
		return false;
	}

	Vertex* vertex = open_set.top();

	if (vertex == end)
	{
		return true;
	}

	closed_set.push_back(vertex);
	open_set.pop();
	
	//Iterate through the neighbour nodes and find the one with the lowest H-Cost
	for (auto v : vertex->neighbours)
	{
		if (!check_closed(v->neighbour))
		{
			v->neighbour->CalculateGCost(v->length, vertex->distance);
			if (!v->neighbour->in_open)
			{
				open_set.push(v->neighbour);
				v->neighbour->in_open = true;
			}
		}
	}

	return compute_route(end);

}

bool AStarGraph::check_closed(Vertex* vertex)
{
	for (auto v : closed_set)
	{
		if (v == vertex)
		{
			return true;
		}
	}

	return false;
}
