//Pathfinding code by Jonathan Robson
#include "astar_graph.h"


#include <chrono>
#include <iostream>
#include <thread>
#include <vector>
#include <list>

// Standard library includes
using std::chrono::duration_cast;
using std::chrono::milliseconds;
using std::cout;
using std::cin;
using std::endl;
using std::this_thread::sleep_for;

// Define the alias "the_clock" for the clock type
typedef std::chrono::steady_clock the_clock;

struct coord
{
	int x, y;
};

const int GRIDSIZE_X = 10;
const int GRIDSIZE_Y = 10;

void ShowGrid(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

void SetGrid(int grid[GRIDSIZE_Y][GRIDSIZE_X]);

void CalculateRoute(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

void CalculateRouteFast(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

bool PhaseOne(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

bool PhaseOneNoWrite(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

void PhaseTwo(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord);

void SpeedTest(int num_nodes, int num_obstacles);

int main(int argc, char *argv[])
{
	bool speed_test = true;
	char test_input;

	while (true)
	{
		cout << "Would you like to run a speed test? (y/n)" << endl;
		cin >> test_input;

		if (test_input != 'y')
		{
			speed_test = false;
		}
		else
		{
			speed_test = true;
		}

		test_input = ' ';
		cout << "Would you like to run A-Star or Lees Pathfindings? (a/l)" << endl;
		cin >> test_input;
		if (test_input == 'a')
		{
			int size = 0;
			int obstacles = 0;

			cout << "\nEnter the number of nodes the graph should consist of: ";
			cin >> size;


			cout << "\nEnter the number of obstacles that should be placed: ";
			cin >> obstacles;

			if (!speed_test)
			{
				AStarGraph graph;

				graph.init_graph(size, obstacles);

				cout << "\nWould you like to see the finished graph? (y/n)" << endl;

				char input;

				cin >> input;

				if (input == 'y')
				{
					graph.display_graph();
				}

				// Start timing
				the_clock::time_point start = the_clock::now();

				//CalculateRoute(grid, start_coord, end_coord);
				graph.test();

				// Stop timing
				the_clock::time_point end = the_clock::now();

				// Compute the difference between the two times in milliseconds
				auto time_taken = duration_cast<milliseconds>(end - start).count();
				cout << "It took " << time_taken << " ms.\n\n";

				graph.clean_up();
			}
			else
			{
				SpeedTest(size, obstacles);
			}
		}
		else if (test_input == 'l')
		{
			if (speed_test)
			{
				float total_time = 0;

				int num_runs = 0;

				cout << "\nEnter the number of runs the program should do: "; cin >> num_runs; cout << "\n";

				for (int i = 0; i < num_runs; i++)
				{
					int grid[GRIDSIZE_Y][GRIDSIZE_X];

					coord start_coord = { 0, 0 };
					coord end_coord = { GRIDSIZE_X - 1, GRIDSIZE_Y - 1 };

					SetGrid(grid);

					// Start timing
					the_clock::time_point start = the_clock::now();

					CalculateRouteFast(grid, start_coord, end_coord);

					// Stop timing
					the_clock::time_point end = the_clock::now();

					// Compute the difference between the two times in milliseconds
					auto time_taken = duration_cast<milliseconds>(end - start).count();

					total_time += time_taken;
				}

				total_time /= num_runs;

				cout << "Average run time was " << total_time << "ms\n";
			}
			else
			{
				float total_time = 0;

				int num_runs = 0;

				int grid[GRIDSIZE_Y][GRIDSIZE_X];

				coord start_coord = { 0, 0 };
				coord end_coord = { GRIDSIZE_X - 1, GRIDSIZE_Y - 1 };

				SetGrid(grid);

				// Start timing
				the_clock::time_point start = the_clock::now();

				CalculateRoute(grid, start_coord, end_coord);

				// Stop timing
				the_clock::time_point end = the_clock::now();

				// Compute the difference between the two times in milliseconds
				auto time_taken = duration_cast<milliseconds>(end - start).count();

				//Show the gird
				ShowGrid(grid, start_coord, end_coord);

			}
		}
	}

	return 0;
}

void SpeedTest(int num_nodes, int num_obstacles)
{
	float total_time = 0;

	int num_runs = 0;

	cout << "\nEnter the number of runs the program should do: "; cin >> num_runs; cout << "\n";

	for (int i = 0; i < num_runs; i++)
	{
		AStarGraph graph;

		graph.init_graph(num_nodes, num_obstacles);

		// Start timing
		the_clock::time_point start = the_clock::now();

		//CalculateRoute(grid, start_coord, end_coord);
		graph.speed_test();

		// Stop timing
		the_clock::time_point end = the_clock::now();

		// Compute the difference between the two times in milliseconds
		auto time_taken = duration_cast<milliseconds>(end - start).count();

		graph.clean_up();

		total_time += time_taken;
	}

	total_time /= num_runs;

	cout << "Average run time was " << total_time << "ms\n";
}

void ShowGrid(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	for (int y = 0; y < GRIDSIZE_Y; y++)
	{
		for (int x = 0; x < GRIDSIZE_X; x++)
		{
			if (x == start_coord.x && y == start_coord.y)
			{
				cout << " " << "A" << " ";
			}
			else if (x == end_coord.x && y == end_coord.y)
			{
				cout << " " << "B" << " ";
			}
			else
			{
				if (grid[y][x] > 9 || grid[y][x] < 0)
				{
					cout << grid[y][x] << " ";
				}
				else
				{
					cout << " " << grid[y][x] << " ";
				}
			}
		}
		cout << endl;
		cout << endl;
	}
}

void SetGrid(int grid[GRIDSIZE_Y][GRIDSIZE_X])
{
	int random;

	srand(time(NULL));

	for (int y = 0; y < GRIDSIZE_Y; y++)
	{
		for (int x = 0; x < GRIDSIZE_X; x++)
		{
			random = rand() % 100;
			if (random >= 80)
			{
				grid[y][x] = -2;
			}
			else
			{
				grid[y][x] = -1;
			}
		}
	}
}

void CalculateRoute(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	if (PhaseOne(grid, start_coord, end_coord))
	{
		PhaseTwo(grid, start_coord, end_coord);
	}
}

void CalculateRouteFast(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	if (PhaseOneNoWrite(grid, start_coord, end_coord))
	{
		PhaseTwo(grid, start_coord, end_coord);
	}
}

bool PhaseOne(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	int numObstacles = 0;

	for (int y = 0; y < GRIDSIZE_Y; y++)
	{
		for (int x = 0; x < GRIDSIZE_X; x++)
		{
			if (grid[y][x] == -2)
			{
				numObstacles++;
			}
		}
	}

	if (start_coord.x < 0 || start_coord.x >= GRIDSIZE_X)
	{
		cout << "The start coordinate is not on the grid." << endl;
		return false;
	}
	else if (start_coord.y < 0 || start_coord.y >= GRIDSIZE_Y)
	{
		cout << "The start coordinate is not on the grid." << endl;
		return false;
	}
	else if (end_coord.x < 0 || start_coord.x >= GRIDSIZE_X)
	{
		cout << "The end coordinate is not on the grid." << endl;
		return false;
	}
	else if (end_coord.y < 0 || start_coord.y >= GRIDSIZE_Y)
	{
		cout << "The end coordinate is not on the grid." << endl;
		return false;
	}

	grid[start_coord.y][start_coord.x] = 0;

	int route_length = 0;

	while (route_length < ((GRIDSIZE_X * GRIDSIZE_Y) - numObstacles))
	{
		for (int y = 0; y < GRIDSIZE_Y; y++)
		{
			for (int x = 0; x < GRIDSIZE_X; x++)
			{
				if (grid[y][x] == route_length)
				{
					if (grid[y - 1][x] == -1 && y - 1 >= 0)
					{
						grid[y - 1][x] = route_length + 1;
						if (y - 1 == end_coord.y && x == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y + 1][x] == -1 && y + 1 < GRIDSIZE_Y)
					{
						grid[y + 1][x] = route_length + 1;
						if (y + 1 == end_coord.y && x == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y][x - 1] == -1 && x - 1 >= 0)
					{
						grid[y][x - 1] = route_length + 1;
						if (y == end_coord.y && x - 1 == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y][x + 1] == -1 && x + 1 < GRIDSIZE_X)
					{
						grid[y][x + 1] = route_length + 1;
						if (y == end_coord.y && x + 1 == end_coord.x)
						{
							return true;
						}
					}
				}
			}
		}
		route_length++;
	}

	cout << "There is no route from the start point to the end point." << endl << endl;
	return false;
}

bool PhaseOneNoWrite(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	int numObstacles = 0;

	for (int y = 0; y < GRIDSIZE_Y; y++)
	{
		for (int x = 0; x < GRIDSIZE_X; x++)
		{
			if (grid[y][x] == -2)
			{
				numObstacles++;
			}
		}
	}

	if (start_coord.x < 0 || start_coord.x >= GRIDSIZE_X)
	{
		return false;
	}
	else if (start_coord.y < 0 || start_coord.y >= GRIDSIZE_Y)
	{
		return false;
	}
	else if (end_coord.x < 0 || start_coord.x >= GRIDSIZE_X)
	{
		return false;
	}
	else if (end_coord.y < 0 || start_coord.y >= GRIDSIZE_Y)
	{
		return false;
	}

	grid[start_coord.y][start_coord.x] = 0;

	int route_length = 0;

	while (route_length < ((GRIDSIZE_X * GRIDSIZE_Y) - numObstacles))
	{
		for (int y = 0; y < GRIDSIZE_Y; y++)
		{
			for (int x = 0; x < GRIDSIZE_X; x++)
			{
				if (grid[y][x] == route_length)
				{
					if (grid[y - 1][x] == -1 && y - 1 >= 0)
					{
						grid[y - 1][x] = route_length + 1;
						if (y - 1 == end_coord.y && x == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y + 1][x] == -1 && y + 1 < GRIDSIZE_Y)
					{
						grid[y + 1][x] = route_length + 1;
						if (y + 1 == end_coord.y && x == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y][x - 1] == -1 && x - 1 >= 0)
					{
						grid[y][x - 1] = route_length + 1;
						if (y == end_coord.y && x - 1 == end_coord.x)
						{
							return true;
						}
					}

					if (grid[y][x + 1] == -1 && x + 1 < GRIDSIZE_X)
					{
						grid[y][x + 1] = route_length + 1;
						if (y == end_coord.y && x + 1 == end_coord.x)
						{
							return true;
						}
					}
				}
			}
		}
		route_length++;
	}
	return false;
}

void PhaseTwo(int grid[GRIDSIZE_Y][GRIDSIZE_X], coord start_coord, coord end_coord)
{
	std::vector<coord> route_list;

	int route_length = grid[end_coord.y][end_coord.x];

	int route_position = route_length;

	coord current_position = { end_coord.x, end_coord.y };

	route_list.push_back(current_position);

	while (route_position > 0)
	{
		if (grid[current_position.y - 1][current_position.x] == route_position - 1 && current_position.y - 1 >= 0)
		{
			route_list.push_back(current_position);
			current_position.x = current_position.x;
			current_position.y = current_position.y - 1;
			route_position--;
		}
		else if (grid[current_position.y + 1][current_position.x] == route_position - 1 && current_position.y + 1 < GRIDSIZE_Y)
		{
			route_list.push_back(current_position);
			current_position.x = current_position.x;
			current_position.y = current_position.y + 1;
			route_position--;
		}
		else if (grid[current_position.y][current_position.x - 1] == route_position - 1 && current_position.x - 1 >= 0)
		{
			route_list.push_back(current_position);
			current_position.x = current_position.x - 1;
			current_position.y = current_position.y;
			route_position--;
		}
		else if (grid[current_position.y][current_position.x + 1] == route_position - 1 && current_position.x + 1 < GRIDSIZE_X)
		{
			route_list.push_back(current_position);
			current_position.x = current_position.x + 1;
			current_position.y = current_position.y;
			route_position--;
		}
	}

	route_list.push_back(start_coord);

	for (int i = 0; i < route_list.size(); i++)
	{
		grid[route_list[i].y][route_list[i].x] = -3;
	}

}